/*********************************************************************

   Copyright (C) 2001 Smaller Animals Software, Inc.

   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
   arising from the use of this software.

   Permission is granted to anyone to use this software for any purpose,
   including commercial applications, and to alter it and redistribute it
   freely, subject to the following restrictions:

   1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.

   2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.

   3. This notice may not be removed or altered from any source distribution.

   http://www.smalleranimals.com
   smallest@smalleranimals.com

   --------

   This code is based, in part, on:

      "Syntax Coloring Text Edit Window Class " - Randy More (9, May 1998)
      http://www.codeguru.com/editctrl/syntax_coloring.shtml

   Though probably only 5% of the original code remains.

**********************************************************************/

/////////////////////////////////////////////////////////////////////////////
//Header for the ColorEditWnd class
/////////////////////////////////////////////////////////////////////////////


#ifndef _COLORIZED_VIEW_
#define _COLORIZED_VIEW_

#include "Colorizer.h"
#include "ScriptText.h"
#include "UndoFrame.h"

class ColorEditWnd : public CWnd
{

public:
   //DECLARE_CONTROL(ColorEditWnd, CWnd);
   DECLARE_DYNAMIC(ColorEditWnd);

   ColorEditWnd();

	ColorEditWnd(CWnd * parent, 
               const CRect & rect,        // where to put it
               int nID,                   // control ID. this control will post messages to the parent
                                          // (EN_CHANGE, EN_KILLFOCUS, EN_SETFOCUS, etc) with this ID

               int iTabSize,              // spaces per tab. all tabs are spaces
               int iFontSize,             // font size in 10ths of a point (uses CreatePointFont)
               CString pFontName,     //RJA changed to CString // font name. only use this with fixed width fonts!!!!!
               bool bShowHorizScroll = true);    // show the horizontal scroll bar?

   
   BOOL Create(DWORD dwFlags,             // recommend : WS_CHILD | WS_VISIBLE | WS_CLIPCHILDREN | WS_VSCROLL | WS_HSCROLL
               CWnd * parent, 
               const CRect & rect,
               int nID, 
               int iTabSize,
               int iFontSize, 
               //const char *pFontName); //RJA
			   CString pFontName);

   // this is how we pick up the cut/copy/paste/undo/redo messages
   //
   // feel free to disable this, if you want to control these functions
   // externally.
   void     UseKeyboardAccelerators(bool b)  {m_bUseKBAccel = b;}

public:
   ///////////////////////////////////////////////////////////////////////////
   // clear all text
	void Reset();

   ///////////////////////////////////////////////////////////////////////////
   // clear and set text to pInText
	void LoadText(const char *pInText);

   // get a single string copy of the contents
	void UnloadText(CStringA &pText) const;

   ///////////////////////////////////////////////////////////////////////////
   // number of text lines in the buffer
   _inline int GetLineCount()      {return GetTextSize();}

   ///////////////////////////////////////////////////////////////////////////
   // position of cursor and anchor. 
   // if these are different positions, the characters between are selected.
   void  SetCursorPos(CPoint pos);
   void  SetAnchorPos(CPoint pos);

   _inline CPoint GetCursorPos()   {return CursorPos();}
   _inline CPoint GetAnchorPos()   {return AnchorPos();}

   // another way to set selection
   void  SetSelection(int charPos1, int line1, int charPos2, int line2);
   void  ReplaceSelText(CString sText);//RJA CString  //RJA unsigned

   // set this line as the top line in the window
   void ScrollToLine(int iLine);

   // scroll if necessary to make this line visible
   void SoftScrollToLine(int iLine);

   // scroll the window to show the cursor
   void ScrollToCursor();

   // get one line of text
   CStringA GetLine(int iLine) const      {return GetTextLine(iLine);}

   // set a single line of text
   void SetLine(const char *pText, int iLine);

   ///////////////////////////////////////////////////////////////////////////
   // client rect of the text at a given line
   CRect GetLineRect(int iLine) const;

   // client rect that encompasses both lines
   CRect GetLineRect(int iLine1, int iLine2) const;

   // number of visible lines. 
   // if bTotal = false, returns the actual number of lines drawn (which may be less
   // than the number of lines that could be drawn, if there was more text...)
   int VisibleLines(bool bTotal = false) const;

   // last and first visible line numbers
   int LastVisibleLine() const;
   int FirstVisibleLine() const;

   // characters on screen
   int VisibleColumns() const;

   ///////////////////////////////////////////////////////////////////////////
   // the parent window can call these to do cut/paste/etc. functions.

   // remove selection to the clipboard
   void DoCut();

   // copy selection to the clipboard
	void DoCopy();

   // paste clipboard. replaces any selection.
	void DoPaste();

   // undo the last action
   void DoUndo();

   // redo the last undo'd action
   void DoRedo();

   // additioinal horizontal spacing
   void SetCharXSpacing(int i);

   // additional vertical spacing
   void SetCharYSpacing(int i);

   // use your own CColorizer class. 
   // this can actually be changed on-the-fly, if you 
   // need to use a different coloring scheme at will.
   void SetColorizer(CColorizer * pColorizer);

   // basic editing stuff
   void SetBackgroundColor(COLORREF clr);
   void SetHighlightColor(COLORREF clr);
   void SetHighlightTextColor(COLORREF clr);

   // make this a read-only control
   void SetReadOnly(bool b)      {m_bReadOnly = b;}

   void ClearUndoHistory();

   // force a recolorization of all text
   void Recolorize();

public:
	virtual ~ColorEditWnd();

private:
   // cursor/anchor position 
   void ClipAnchorPosition(int &x, int &y);
   void ClipCursorPosition(int &x, int &y);
   void ClipAnchorPosition(CPoint &p);
   void ClipCursorPosition(CPoint &p);
   
   // warning: these only affect the internal state. they do not update the UI!
   _inline void _SetAnchorPosX(int x)      {m_anchorPos.x = x;}
   _inline void _SetAnchorPosY(int y)      {m_anchorPos.y = y;}
   _inline void _SetAnchorPos(CPoint pt)   {m_anchorPos = pt;}
   _inline void _SetAnchorPos(int x, int y) {m_anchorPos.x = x; m_anchorPos.y = y;}

   // warning: these only set the internal data. they do not update the UI!
   _inline void _SetCursorPosX(int x)      {m_cursorPos.x = x;}
   _inline void _SetCursorPosY(int y)      {m_cursorPos.y = y;}
   _inline void _SetCursorPos(CPoint pt)   {m_cursorPos = pt;}
   _inline void _SetCursorPos(int x, int y) {m_cursorPos.x= x; m_cursorPos.y = y;}

   _inline CPoint   CursorPos() const     {return m_cursorPos;}
   _inline int   CursorPosX() const       {return m_cursorPos.x;}
   _inline int   CursorPosY() const       {return m_cursorPos.y;}

   _inline CPoint   AnchorPos() const     {return m_anchorPos;}
   _inline int   AnchorPosX() const       {return m_anchorPos.x;}
   _inline int   AnchorPosY() const       {return m_anchorPos.y;}

   // number of lines
   _inline int GetTextSize() const        {return m_text._GetSize();}

   // get a single line
   _inline CStringA GetTextLine(int i) const       {return m_text._GetLine(i);}

   // insert/delete

   // insert a single line at pos 'i'. 
   void InsertTextLine(CStringA &line, 
                        int i, 
                        bool bUpdate=TRUE,      // redraw and recolor. in general, TRUE
                        bool bNoUndo = false);  // don't add to undo frame

   // insert a string at the current cursor pos
   void InsertText(CStringA &pString);

   // insert a single char at the current cursor pos
	void InsertChar(int pChar, bool pUpdate = TRUE);

   // insert a string at the current cursor pos
	void InsertString(CStringA &pString, bool pUpdate = TRUE);

   // insert a single tab (as spaces) at the current cursor pos
   void  InsertTab();

   // set a single text line. replaces old contents
   void SetTextLine(CStringA &line, int i, bool bNoUndo = false, bool pUpdate = TRUE);

   // remove the line at 'i'
   void RemoveTextLine(int i, bool bNoUndo = false, bool pUpdate = TRUE);                     

   // clear all lines
   void RemoveAllText();                           

   void  _ReplaceSelText(CString sText);//RJA changing to CString due to code page dropping circuit characters copied from prop tool  //RJA changing to unsigned for extended ASCII support

   ///////////////////////////////////////////////////////////////////////////

private:
   CStringA MakeTab(int iCursorXPos) const;

   // for the backspace and delete keys
	void DeleteNextChar();
   void DeletePrevChar();

   // handle enter key
   void DoEnter();

   // selection
	bool IsSelected(int pX,int pY);
	bool Selected();
	void UpdateAnchor(bool pTestShift);
	void RemoveSelected();

   // scroll
   void UpdateScrollSizes();
   
   CPoint GetDeviceScrollPosition() const;
   CPoint GetScrollPosition() const;

   // output
   CPoint _DrawLine(CDC *pDC, CStringA &line, const CScriptText::lineVector &colors, int lineidx, int y);
   void DrawCursor(CDC * pDC, bool pShow);
	void DrawCursor(bool pShow);

   void ShiftDisplay(int dx, int dy);


   // undo stuff
   void StartUndoFrame();
   void EndUndoFrame();
   void ClearUndoStack(undostack &inStack);
   void DoUndoRedo(undostack &inStack, undostack& outStack);

   // start/end an undoable action
   void StartAction();
   void EndAction();

   // post a windows message to the parent
   void  NotifyParent(UINT msg);

   // state
	int m_iCharDX;
	int m_iCharDY;
	int m_iTabSize;
	HCURSOR m_cursor;
   CFont m_font;
   CFont m_fontBold;  //RJA:  Need bold font for keywords

	// offsets for scrolling
	long			m_iFirstCharacter;
	long			m_iFirstLine;

	//caret and cursor postion
   CPoint      m_anchorPos;
   CPoint      m_cursorPos;

   // colors
   COLORREF m_backgroundColor;
   COLORREF m_hilightColor;
   COLORREF m_hilightTextColor;

   // the text
   CScriptText		m_text;

   // sel/drag info
	bool     m_bIsSelected;
	bool     m_bLBDown;
	bool     m_bCaretOn;
   CPoint   m_startPointDrag;

   // use accellerators?
   bool     m_bUseKBAccel;

   bool     m_bMaintainIndent;


   // undo/redo state

   // current action frame. an action can be composed of one or more sub-actions.
   // for a single keystroke, the undo frame holds one action: the keystroke.
   // for complex actions, like a paste or a cut, the frame will contain all of the 
   // actions required to perform the meta-action.
   CUndoFrame  m_currentUndoFrame;

   // holds all of our actions - this might get big. feel free to clear it once in a while.
   undostack   m_undoStack;

   // when the user starts undo-ing stuff, the undo's are placed here. 
   // this is cleared when a new undo frame is created, ie. when a new edit action occurs.
   undostack   m_redoStack;

   // dialog ID
   UINT  m_ID; 

   bool  m_bReadOnly;

   HACCEL m_hAccel;

public:
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(ColorEditWnd)
	public:
	virtual BOOL PreTranslateMessage(MSG* pMsg);
	//}}AFX_VIRTUAL

	//{{AFX_MSG(ColorEditWnd)
	afx_msg BOOL OnEraseBkgnd(CDC* pDC);
	afx_msg void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
   afx_msg void OnChar(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnLButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnLButtonUp(UINT nFlags, CPoint point);
	afx_msg void OnMouseMove(UINT nFlags, CPoint point);
	afx_msg void OnPaint();
	afx_msg void OnSize(UINT nType, int cx, int cy);
	afx_msg UINT OnGetDlgCode();
	afx_msg void OnVScroll( UINT nSBCode, UINT nPos, CScrollBar* pScrollBar );
	afx_msg void OnHScroll( UINT nSBCode, UINT nPos, CScrollBar* pScrollBar );
	afx_msg int OnCreate(LPCREATESTRUCT lpCreateStruct);
	afx_msg BOOL OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT message);
	afx_msg void OnShowWindow(BOOL bShow, UINT nStatus);
   afx_msg BOOL OnMouseWheel( UINT nFlags, short zDelta, CPoint pt );
	afx_msg void OnContextMenu(CWnd* pWnd, CPoint point);
	afx_msg void OnUpdateCopy(CCmdUI* pCmdUI);
	afx_msg void OnUpdateCut(CCmdUI* pCmdUI);
	afx_msg void OnUpdatePaste(CCmdUI* pCmdUI);
	afx_msg void OnUpdateSelectall(CCmdUI* pCmdUI);
	afx_msg void OnUpdateUndo(CCmdUI* pCmdUI);
   afx_msg void OnUpdateRedo(CCmdUI* pCmdUI);
	afx_msg void OnCopy();
	afx_msg void OnCut();
	afx_msg void OnPaste();
	afx_msg void OnSelectall();
	afx_msg void OnUndo();
	afx_msg void OnRedo();
   afx_msg void OnKillFocus(CWnd* pNewWnd);
   afx_msg void OnSetFocus( CWnd* pOldWnd );
   afx_msg void OnLButtonDblClk( UINT, CPoint );
	//}}AFX_MSG
   DECLARE_MESSAGE_MAP()

public:
   CStringA m_className;

   //RJA Moving to ScriptText 

   //Code page
   TCHAR* m_d[256] = { L"\xf000\x00",L"\xf001\x00",L"\x2190\x00",L"\x2192\x00",L"\x2191\x00",L"\x2193\x00",L"\x25C0\x00",L"\x25B6\x00",L"\xF008\x00",L"\xF009\x00",L"\xF00a\x00",L"\xf00b\x00",L"\xf00c\x00",L"\xf00d\x00",L"\x2023\x00",L"\x2022\x00",
	   L"\x0394\x00",L"\x03C0\x00",L"\x03A3\x00",L"\x03a9\x00",L"\x2248\x00",L"\x221a\x00",L"\xF016\x00",L"\xF017\x00",L"\xF018\x00",L"\xF019\x00",L"\xF01a\x00",L"\xF01b\x00",L"\xF01c\x00",L"\xF01d\x00",L"\xF01e\x00",L"\xF01f\x00",

	   L"\x20\x00",L"\x21\x00",L"\x22\x00",L"\x23\x00",L"\x24\x00",L"\x25\x00",L"\x26\x00",L"\x27\x00",L"\x28\x00",L"\x29\x00",L"\x2a\x00",L"\x2b\x00",L"\x2c\x00",L"\x2d\x00",L"\x2e\x00",L"\x2f\x00",
	   L"\x30\x00",L"\x31\x00",L"\x32\x00",L"\x33\x00",L"\x34\x00",L"\x35\x00",L"\x36\x00",L"\x37\x00",L"\x38\x00",L"\x39\x00",L"\x3a\x00",L"\x3b\x00",L"\x3c\x00",L"\x3d\x00",L"\x3e\x00",L"\x3f\x00",

	   L"\x40\x00",L"\x41\x00",L"\x42\x00",L"\x43\x00",L"\x44\x00",L"\x45\x00",L"\x46\x00",L"\x47\x00",L"\x48\x00",L"\x49\x00",L"\x4a\x00",L"\x4b\x00",L"\x4c\x00",L"\x4d\x00",L"\x4e\x00",L"\x4f\x00",
	   L"\x50\x00",L"\x51\x00",L"\x52\x00",L"\x53\x00",L"\x54\x00",L"\x55\x00",L"\x56\x00",L"\x57\x00",L"\x58\x00",L"\x59\x00",L"\x5a\x00",L"\x5b\x00",L"\x5c\x00",L"\x5d\x00",L"\x5e\x00",L"\x5f\x00",

	   L"\x60\x00",L"\x61\x00",L"\x62\x00",L"\x63\x00",L"\x64\x00",L"\x65\x00",L"\x66\x00",L"\x67\x00",L"\x68\x00",L"\x69\x00",L"\x6a\x00",L"\x6b\x00",L"\x6c\x00",L"\x6d\x00",L"\x6e\x00",L"\x6f\x00",
	   L"\x70\x00",L"\x71\x00",L"\x72\x00",L"\x73\x00",L"\x74\x00",L"\x75\x00",L"\x76\x00",L"\x77\x00",L"\x78\x00",L"\x79\x00",L"\x7a\x00",L"\x7b\x00",L"\x7c\x00",L"\x7d\x00",L"\x7e\x00",L"\xF07f\x00",

	   L"\xf080\x00",L"\xf081\x00",L"\xf082\x00",L"\xf083\x00",L"\xf084\x00",L"\xf085\x00",L"\xf086\x00",L"\xf087\x00",L"\xf088\x00",L"\xf089\x00",L"\xf08a\x00",L"\xf08b\x00",L"\xf08c\x00",L"\xf08d\x00",L"\xf08e\x00",L"\xf08f\x00",
	   L"\x2500\x00",L"\x2502\x00",L"\x253C\x00",L"\x254b\x00",L"\x2524\x00",L"\x251c\x00",L"\x2534\x00",L"\x252c\x00",L"\x252b\x00",L"\x2523\x00",L"\x253b\x00",L"\x2533\x00",L"\x2518\x00",L"\x2514\x00",L"\x2510\x00",L"\x250c\x00",

	   L"\xf0a0\x00",L"\x00a1\x00",L"\xf0a2\x00",L"\x00a3\x00",L"\x20ac\x00",L"\x00a5\x00",L"\xf0a6\x00",L"\xf0a7\x00",L"\xf0a8\x00",L"\xf0a9\x00",L"\xf0aa\x00",L"\xf0ab\x00",L"\xf0ac\x00",L"\xf0ad\x00",L"\xf0ae\x00",L"\xf0af\x00",
	   L"\x00b0\x00",L"\x00b1\x00",L"\x00b2\x00",L"\x00b3\x00",L"\xf0b4\x00",L"\x00b5\x00",L"\xf0b6\x00",L"\xf0b7\x00",L"\xf0b8\x00",L"\x00b9\x00",L"\xf0ba\x00",L"\xf0bb\x00",L"\xf0bc\x00",L"\xf0bd\x00",L"\xf0be\x00",L"\x00bf\x00",

	   L"\xc0\x00",L"\xc1\x00",L"\xc2\x00",L"\xc3\x00",L"\xc4\x00",L"\xc5\x00",L"\xc6\x00",L"\xc7\x00",L"\xc8\x00",L"\xc9\x00",L"\xca\x00",L"\xcb\x00",L"\xcc\x00",L"\xcd\x00",L"\xce\x00",L"\xcf\x00",
	   L"\xd0\x00",L"\xd1\x00",L"\xd2\x00",L"\xd3\x00",L"\xd4\x00",L"\xd5\x00",L"\xd6\x00",L"\xd7\x00",L"\xd8\x00",L"\xd9\x00",L"\xda\x00",L"\xdb\x00",L"\xdc\x00",L"\xdd\x00",L"\xde\x00",L"\xdf\x00",

	   L"\xe0\x00",L"\xe1\x00",L"\xe2\x00",L"\xe3\x00",L"\xe4\x00",L"\xe5\x00",L"\xe6\x00",L"\xe7\x00",L"\xe8\x00",L"\xe9\x00",L"\xea\x00",L"\xeb\x00",L"\xec\x00",L"\xed\x00",L"\xee\x00",L"\xef\x00",
	   L"\xf0\x00",L"\xf1\x00",L"\xf2\x00",L"\xf3\x00",L"\xf4\x00",L"\xf5\x00",L"\xf6\x00",L"\xf7\x00",L"\xf8\x00",L"\xf9\x00",L"\xfa\x00",L"\xfb\x00",L"\xfc\x00",L"\xfd\x00",L"\xfe\x00",L"\x221e\x00" };

   //Font stuff
   int m_iFontSize;
   CString m_pFontName;

};


/////////////////////////////////////////////////////////////////////////////
#endif
